import decimal
import itertools
import os
import string
import re
import uuid

import attr
import jinja2
import toolz

import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.cantosym
import epcpm.sunspecmodel
import epcpm.sunspectoxlsx

builders = epyqlib.utils.general.TypeMap()


# TODO: move this somewhere common in python code...
sunspec_types = {
    "uint16": "sunsU16",
    "enum16": "sunsU16",
    "int16": "sunsS16",
    "uint32": "sunsU32",
    "int32": "sunsS32",
    "string": "PackedString",
    "bitfield16": "sunsU16",
    "bitfield32": "sunsU32",
}


def node_path_string(node):
    nodes = [node, *node.ancestors()][:-1]
    names = [node.name for node in reversed(nodes)]

    return " > ".join(names)


class InvalidAccessLevelError(Exception):
    @classmethod
    def build(cls, value, parameter):
        message = (
            f"Invalid access level specified for"
            f" {parameter.name} ({parameter.uuid}): {value}"
        )

        return cls(message)


def export(
    c_path,
    h_path,
    parameters_model,
    can_model,
    sunspec_model,
    skip_sunspec=False,
    include_uuid_in_item=False,
):
    if skip_sunspec:
        sunspec_root = None
    else:
        sunspec_root = sunspec_model.root

    builder = builders.wrap(
        wrapped=parameters_model.root,
        can_root=can_model.root,
        sunspec_root=sunspec_root,
        include_uuid_in_item=include_uuid_in_item,
    )

    c_path.parent.mkdir(parents=True, exist_ok=True)

    built_c, built_h, model_ids = builder.gen()

    template_context = {
        "sunspec_interface_gen_headers": (
            f"sunspecInterfaceGen{id}.h" for id in model_ids
        ),
        "sunspec_interface_headers": (f"sunspecInterface{id:05}.h" for id in model_ids),
        "interface_items": epcpm.c.format_nested_lists(
            built_c,
        ).strip(),
        "declarations": epcpm.c.format_nested_lists(built_h).strip(),
    }

    epcpm.c.render(
        source=c_path.with_suffix(f"{c_path.suffix}_pm"),
        destination=c_path,
        context=template_context,
    )

    epcpm.c.render(
        source=h_path.with_suffix(f"{h_path.suffix}_pm"),
        destination=h_path,
        context=template_context,
    )


@builders(epyqlib.pm.parametermodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()

    def gen(self):
        def can_node_wanted(node):
            if getattr(node, "parameter_uuid", None) is None:
                return False

            uuids = [
                # CCP Response
                uuid.UUID("39315d58-1ddb-48b9-960c-96e724c89da1"),
                # CCP
                uuid.UUID("983bdc5d-8d4e-4107-a0a0-983f0ab101ce"),
            ]
            return not any(ancestor.uuid in uuids for ancestor in node.ancestors())

        can_nodes_with_parameter_uuid = self.can_root.nodes_by_filter(
            filter=can_node_wanted,
        )

        parameter_uuid_to_can_node = {
            node.parameter_uuid: node for node in can_nodes_with_parameter_uuid
        }

        def sunspec_node_wanted(node):
            if getattr(node, "parameter_uuid", None) is None:
                return False

            wanted_types = (
                epcpm.sunspecmodel.DataPoint,
                epcpm.sunspecmodel.DataPointBitfieldMember,
            )
            if not isinstance(node, wanted_types):
                return False

            return True

        if self.sunspec_root is None:
            parameter_uuid_to_sunspec_node = {}
        else:
            sunspec_nodes_with_parameter_uuid = self.sunspec_root.nodes_by_filter(
                filter=sunspec_node_wanted,
            )

            parameter_uuid_to_sunspec_node = {
                node.parameter_uuid: node for node in sunspec_nodes_with_parameter_uuid
            }

        lengths_equal = len(can_nodes_with_parameter_uuid) == len(
            parameter_uuid_to_can_node
        )
        if not lengths_equal:
            uuids = [u.parameter_uuid for u in can_nodes_with_parameter_uuid]
            print("\n".join(set(str(u) for u in uuids if uuids.count(u) > 1)))
            raise Exception()

        c = []
        h = []
        sunspec_models = set()

        for child in self.wrapped.children:
            if not isinstance(
                child,
                (
                    epyqlib.pm.parametermodel.Group,
                    epyqlib.pm.parametermodel.Parameter,
                    epyqlib.pm.parametermodel.Table,
                    # epcpm.parametermodel.EnumeratedParameter,
                ),
            ):
                continue

            c_built, h_built, sunspec_models_built = builders.wrap(
                wrapped=child,
                can_root=self.can_root,
                sunspec_root=self.sunspec_root,
                include_uuid_in_item=self.include_uuid_in_item,
                parameter_uuid_to_can_node=parameter_uuid_to_can_node,
                parameter_uuid_to_sunspec_node=(parameter_uuid_to_sunspec_node),
                parameter_uuid_finder=self.wrapped.model.node_from_uuid,
            ).gen()

            c.extend(c_built)
            h.extend(h_built)
            sunspec_models |= sunspec_models_built

        return c, h, sunspec_models

        # return itertools.chain.from_iterable(
        #     builders.wrap(
        #         wrapped=child,
        #         can_root=self.can_root,
        #         sunspec_root=self.sunspec_root,
        #         parameter_uuid_to_can_node=parameter_uuid_to_can_node,
        #         parameter_uuid_to_sunspec_node=(
        #             parameter_uuid_to_sunspec_node
        #         ),
        #         parameter_uuid_finder=self.wrapped.model.node_from_uuid,
        #     ).gen()
        #     for child in parameters.children
        #     if isinstance(
        #         child,
        #         (
        #             epyqlib.pm.parametermodel.Group,
        #             epyqlib.pm.parametermodel.Parameter,
        #             # epcpm.parametermodel.EnumeratedParameter,
        #         ),
        #     )
        # )


@builders(epyqlib.pm.parametermodel.Group)
@attr.s
class Group:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        c = []
        h = []
        sunspec_models = set()

        for child in self.wrapped.children:
            if not isinstance(
                child,
                (
                    epyqlib.pm.parametermodel.Group,
                    epyqlib.pm.parametermodel.Parameter,
                    epyqlib.pm.parametermodel.Table,
                    # epcpm.parametermodel.EnumeratedParameter,
                ),
            ):
                continue

            c_built, h_built, sunspec_models_built = builders.wrap(
                wrapped=child,
                can_root=self.can_root,
                sunspec_root=self.sunspec_root,
                include_uuid_in_item=self.include_uuid_in_item,
                parameter_uuid_to_can_node=(self.parameter_uuid_to_can_node),
                parameter_uuid_to_sunspec_node=(self.parameter_uuid_to_sunspec_node),
                parameter_uuid_finder=self.parameter_uuid_finder,
            ).gen()

            c.extend(c_built)
            h.extend(h_built)
            sunspec_models |= sunspec_models_built

        return c, h, sunspec_models
        # return itertools.chain.from_iterable(
        #     result
        #     for result in (
        #         builders.wrap(
        #             wrapped=child,
        #             can_root=self.can_root,
        #             sunspec_root=self.sunspec_root,
        #             parameter_uuid_to_can_node=(
        #                 self.parameter_uuid_to_can_node
        #             ),
        #             parameter_uuid_to_sunspec_node=(
        #                 self.parameter_uuid_to_sunspec_node
        #             ),
        #             parameter_uuid_finder=self.parameter_uuid_finder,
        #         ).gen()
        #         for child in self.wrapped.children
        #         if isinstance(
        #             child,
        #             (
        #                 epyqlib.pm.parametermodel.Group,
        #                 epyqlib.pm.parametermodel.Parameter,
        #                 # epcpm.parametermodel.EnumeratedParameter,
        #             ),
        #         )
        #     )
        # )


@builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class DataPoint:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    def interface_variable_name(self):
        parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

        maybe_model = self.wrapped.tree_parent

        while not isinstance(maybe_model, epcpm.sunspecmodel.Model):
            maybe_model = maybe_model.tree_parent

        model = maybe_model
        model_variable = f"sunspecInterface.model{model.id}"

        return f"&{model_variable}.{parameter.abbreviation}"


@builders(epcpm.sunspecmodel.DataPointBitfieldMember)
@attr.s
class DataPointBitfieldMember:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    def interface_variable_name(self):
        parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

        uuid_ = str(parameter.uuid).replace("-", "_")
        return f"&interfaceItem_variable_{uuid_}"


@builders(epyqlib.pm.parametermodel.Parameter)
@attr.s
class Parameter:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        parameter = self.wrapped
        can_signal = self.parameter_uuid_to_can_node.get(parameter.uuid)
        sunspec_point = self.parameter_uuid_to_sunspec_node.get(parameter.uuid)

        interface_data = [
            can_signal,
            sunspec_point,
        ]

        uses_interface_item = (
            isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
            and parameter.uses_interface_item()
        )

        sunspec_models = set()

        if not uses_interface_item or all(x is None for x in interface_data):
            return [[], [], sunspec_models]

        scale_factor_variable = "NULL"
        scale_factor_updater = "NULL"

        if parameter.getter_function is None:
            getter_function = "NULL"
        else:
            getter_function = parameter.getter_function

        if parameter.setter_function is None:
            setter_function = "NULL"
        else:
            setter_function = parameter.setter_function

        if parameter.internal_variable is not None:
            var_or_func = "variable"

            variable_or_getter_setter = [
                f".variable = &{parameter.internal_variable},",
            ]
        else:
            var_or_func = "functions"

            variable_or_getter_setter = [
                f".getter = {getter_function},",
            ]

        variable_or_getter_setter.append(f".setter = {setter_function},")

        if sunspec_point is None:
            sunspec_variable = "NULL"
            sunspec_getter = "NULL"
            sunspec_setter = "NULL"
            hand_coded_sunspec_getter_function = "NULL"
            hand_coded_sunspec_setter_function = "NULL"
        else:
            maybe_model = sunspec_point.tree_parent

            while not isinstance(maybe_model, epcpm.sunspecmodel.Model):
                maybe_model = maybe_model.tree_parent

            model = maybe_model

            sunspec_models.add(model.id)

            # TODO: move this somewhere common in python code...
            sunspec_type = sunspec_types[
                self.parameter_uuid_finder(sunspec_point.type_uuid).name
            ]

            # TODO: handle tables with repeating blocks and references

            hand_coded_getter_function_name = epcpm.sunspectoxlsx.getter_name(
                parameter=parameter,
                model_id=model.id,
                is_table=False,
            )

            hand_coded_setter_function_name = epcpm.sunspectoxlsx.setter_name(
                parameter=parameter,
                model_id=model.id,
                is_table=False,
            )

            if getattr(sunspec_point, "hand_coded_getter", False):
                hand_coded_sunspec_getter_function = (
                    f"&{hand_coded_getter_function_name}"
                )
            else:
                hand_coded_sunspec_getter_function = "NULL"

            if getattr(sunspec_point, "hand_coded_setter", False):
                hand_coded_sunspec_setter_function = (
                    f"&{hand_coded_setter_function_name}"
                )
            else:
                hand_coded_sunspec_setter_function = "NULL"

            # TODO: CAMPid 67549654267913467967436
            if getattr(sunspec_point, "factor_uuid", False):
                factor_point = self.sunspec_root.model.node_from_uuid(
                    sunspec_point.factor_uuid,
                )
                sunspec_scale_factor = self.parameter_uuid_finder(
                    factor_point.parameter_uuid,
                ).abbreviation

                sunspec_factor_builder = builders.wrap(
                    wrapped=factor_point,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                )
                scale_factor_variable = sunspec_factor_builder.interface_variable_name()
                scale_factor_updater_name = (
                    f"getSUNSPEC_MODEL{model.id}_{sunspec_scale_factor}"
                )
                scale_factor_updater = f"&{scale_factor_updater_name}"

            sunspec_point_builder = builders.wrap(
                wrapped=sunspec_point,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            sunspec_variable = sunspec_point_builder.interface_variable_name()

            # TODO: CAMPid 9675436715674367943196954756419543975314
            getter_setter_list = [
                "InterfaceItem",
                var_or_func,
                types[parameter.internal_type].name,
                "sunspec",
                sunspec_type,
            ]

            sunspec_getter = "_".join(str(x) for x in getter_setter_list + ["getter"])
            sunspec_setter = "_".join(str(x) for x in getter_setter_list + ["setter"])

        interface_item_type = (
            f"InterfaceItem_{var_or_func}_{types[parameter.internal_type].name}"
        )

        can_getter, can_setter, can_variable = can_getter_setter_variable(
            can_signal=can_signal,
            parameter=parameter,
            var_or_func_or_table=var_or_func,
        )

        access_level = get_access_level_string(
            parameter=parameter,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        if parameter.rejected_callback is None:
            rejected_callback = "NULL"
        else:
            rejected_callback = f"&{parameter.rejected_callback}"

        result = create_item(
            item_uuid=parameter.uuid,
            include_uuid_in_item=self.include_uuid_in_item,
            access_level=access_level,
            can_getter=can_getter,
            can_setter=can_setter,
            can_variable=can_variable,
            hand_coded_sunspec_getter_function=hand_coded_sunspec_getter_function,
            hand_coded_sunspec_setter_function=hand_coded_sunspec_setter_function,
            interface_item_type=interface_item_type,
            internal_scale=parameter.internal_scale_factor,
            meta_initializer_values=create_meta_initializer_values(parameter),
            parameter=parameter,
            scale_factor_updater=scale_factor_updater,
            scale_factor_variable=scale_factor_variable,
            sunspec_getter=sunspec_getter,
            sunspec_setter=sunspec_setter,
            sunspec_variable=sunspec_variable,
            variable_or_getter_setter=variable_or_getter_setter,
            rejected_callback=rejected_callback,
            can_scale_factor=getattr(can_signal, "factor", None),
            reject_from_inactive_interfaces=parameter.reject_from_inactive_interfaces,
        )

        return [*result, sunspec_models]


@attr.s(frozen=True)
class FixedWidthType:
    name = attr.ib()
    type = attr.ib()
    bits = attr.ib()
    signed = attr.ib()
    minimum_code = attr.ib()
    maximum_code = attr.ib()

    @classmethod
    def build(cls, bits, signed):
        return cls(
            name=fixed_width_name(bits=bits, signed=signed),
            type=fixed_width_name(bits=bits, signed=signed),
            bits=bits,
            signed=signed,
            minimum_code=fixed_width_limit_text(
                bits=bits,
                signed=signed,
                limit="min",
            ),
            maximum_code=fixed_width_limit_text(
                bits=bits,
                signed=signed,
                limit="max",
            ),
        )


@attr.s(frozen=True)
class FloatingType:
    name = attr.ib()
    type = attr.ib()
    bits = attr.ib()
    minimum_code = attr.ib()
    maximum_code = attr.ib()

    @classmethod
    def build(cls, bits):
        return cls(
            name={32: "float", 64: "double"}[bits],
            type={32: "float", 64: "double"}[bits],
            bits=bits,
            minimum_code="(-INFINITY)",
            maximum_code="(INFINITY)",
        )


@attr.s(frozen=True)
class BooleanType:
    name = attr.ib(default="bool")
    type = attr.ib(default="bool")
    bits = attr.ib(default=2)
    minimum_code = attr.ib(default="(false)")
    maximum_code = attr.ib(default="(true)")


@attr.s(frozen=True)
class SizeType:
    name = attr.ib(default="size_t")
    type = attr.ib(default="size_t")
    bits = attr.ib(default=32)
    minimum_code = attr.ib(default="(0)")
    maximum_code = attr.ib(default="(SIZE_MAX)")


@attr.s(frozen=True)
class VoidPointerType:
    name = attr.ib(default="void_p")
    type = attr.ib(default="void*")
    minimum_code = attr.ib(default="((void *) 0)")
    maximum_code = attr.ib(default="((void *) UINT32_C(0x240000))")


@attr.s(frozen=True)
class PackedStringType:
    name = attr.ib(default="PackedString")
    type = attr.ib(default="PackedString")
    minimum_code = attr.ib(default="(0)")
    maximum_code = attr.ib(default="(0)")


def fixed_width_name(bits, signed):
    if signed:
        u = ""
    else:
        u = "u"

    return f"{u}int{bits}_t"


def fixed_width_limit_text(bits, signed, limit):
    limits = ("min", "max")

    if limit not in limits:
        raise Exception(f"Requested limit not found in {list(limits)}: {limit:!r}")

    if not signed and limit == "min":
        return "(0U)"

    u = "" if signed else "U"

    return f"({u}INT{bits}_{limit.upper()})"


types = {
    type.type: type
    for type in (
        *(
            FixedWidthType.build(
                bits=bits,
                signed=signed,
            )
            for bits in (8, 16, 32, 64)
            for signed in (False, True)
        ),
        *(FloatingType.build(bits=bits) for bits in (32, 64)),
        BooleanType(),
        SizeType(),
        VoidPointerType(),
        PackedStringType(),
    )
}


def create_meta_initializer_values(parameter):
    def create_literal(value, type):
        value *= decimal.Decimal(10) ** parameter.internal_scale_factor

        suffix = ""

        if type == "float":
            suffix = "f"
            value = float(value)
        elif type == "bool":
            value = str(bool(value)).lower()
        elif type.startswith("uint"):
            suffix = "U"
            value = int(round(value))
        else:
            value = int(round(value))

        return str(value) + suffix

    if parameter.default is None:
        meta_default = 0
    else:
        meta_default = parameter.default
    meta_default = create_literal(
        value=meta_default,
        type=parameter.internal_type,
    )

    if parameter.minimum is None:
        meta_minimum = types[parameter.internal_type].minimum_code
    else:
        meta_minimum = parameter.minimum
        meta_minimum = create_literal(
            value=meta_minimum,
            type=parameter.internal_type,
        )

    if parameter.maximum is None:
        meta_maximum = types[parameter.internal_type].maximum_code
    else:
        meta_maximum = parameter.maximum
        meta_maximum = create_literal(
            value=meta_maximum,
            type=parameter.internal_type,
        )

    meta_initializer_values = [
        f"[Meta_UserDefault - 1] = {meta_default},",
        f"[Meta_FactoryDefault - 1] = {meta_default},",
        f"[Meta_Min - 1] = {meta_minimum},",
        f"[Meta_Max - 1] = {meta_maximum}",
    ]
    return meta_initializer_values


def get_access_level_string(parameter, parameter_uuid_finder):
    access_level_uuid = parameter.access_level_uuid

    try:
        access_level = parameter_uuid_finder(access_level_uuid)
    except epyqlib.attrsmodel.NotFoundError as e:
        raise InvalidAccessLevelError.build(
            value=access_level_uuid,
            parameter=parameter,
        ) from e

    access_level = f"CAN_Enum_AccessLevel_{access_level.name}"

    return access_level


def can_getter_setter_variable(can_signal, parameter, var_or_func_or_table):
    if can_signal is None:
        can_variable = "NULL"
        can_getter = "NULL"
        can_setter = "NULL"

        return can_getter, can_setter, can_variable

    in_table = isinstance(
        can_signal.tree_parent.tree_parent,
        epcpm.canmodel.CanTable,
    )
    if in_table:
        can_variable = (
            f"&{can_signal.tree_parent.tree_parent.tree_parent.name}"
            f".{can_signal.tree_parent.tree_parent.name}"
            f"{can_signal.tree_parent.name}"
            f".{can_signal.name}"
        )
    else:
        can_variable = (
            f"&{can_signal.tree_parent.tree_parent.name}"
            f".{can_signal.tree_parent.name}"
            f".{can_signal.name}"
        )

    if can_signal.signed:
        can_type = ""
    else:
        can_type = "u"

    can_type += "int"

    if can_signal.bits <= 16:
        can_type += "16"
    elif can_signal.bits <= 32:
        can_type += "32"
    else:
        raise Exception("ack")

    can_type += "_t"

    getter_setter_list = [
        "InterfaceItem",
        var_or_func_or_table,
        types[parameter.internal_type].name,
        "can",
        can_type,
    ]

    can_getter = "_".join(str(x) for x in getter_setter_list + ["getter"])
    can_setter = "_".join(str(x) for x in getter_setter_list + ["setter"])

    return can_getter, can_setter, can_variable


# TODO: CAMPid 68945967541316743769675426795146379678431
def breakdown_nested_array(s):
    split = re.split(r"\[(.*?)\].", s)

    array_layers = list(toolz.partition(2, split))
    (remainder,) = split[2 * len(array_layers) :]

    return array_layers, remainder


# TODO: CAMPid 0974567213671436714671907842679364
@attr.s
class NestedArrays:
    array_layers = attr.ib()
    remainder = attr.ib()

    @classmethod
    def build(cls, s):
        array_layers, remainder = breakdown_nested_array(s)

        return cls(
            array_layers=array_layers,
            remainder=remainder,
        )

    def index(self, indexes):
        try:
            return ".".join(
                "{layer}[{index}]".format(
                    layer=layer,
                    index=index_format.format(**indexes),
                )
                for (layer, index_format), index in zip(self.array_layers, indexes)
            )
        except KeyError as e:
            raise

    def sizeof(self, layers):
        indexed = self.index(indexes={layer: 0 for layer in layers})

        return f"sizeof({indexed})"

    # def sizeof(self, layers, remainder=False):
    #     indexed = self.index(indexes={layer: 0 for layer in layers})

    #     if remainder:
    #         if len(layers) != len(self.array_layers):
    #             raise Exception('Remainder requested without specifying all layers')

    #         indexed += f'.{self.remainder}'

    #     return f'sizeof({indexed})'


@attr.s
class TableBaseStructures:
    array_nests = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()
    include_uuid_in_item = attr.ib()
    common_structure_names = attr.ib(factory=dict)
    c_code = attr.ib(factory=list)
    h_code = attr.ib(factory=list)

    def ensure_common_structure(
        self,
        internal_type,
        internal_name,
        parameter,
        remainder,
        common_initializers,
        meta_initializer,
        setter,
    ):
        name = self.common_structure_names.get(parameter.uuid)

        if name is None:
            if len(self.h_code) > 0:
                self.h_code.append("")
            if len(self.c_code) > 0:
                self.c_code.append("")

            formatted_uuid = str(parameter.uuid).replace("-", "_")
            name = f"InterfaceItem_table_common_{internal_type}_{formatted_uuid}"

            if remainder is None:
                sizes = {}
                full_base_variable = "NULL"
            else:
                nested_array = self.array_nests[remainder]

                layers = []
                for layer in nested_array.array_layers:
                    (layer_format_name,) = [
                        list(field)[0][1]
                        for field in [string.Formatter().parse(layer[1])]
                    ]
                    layers.append(layer_format_name)

                variable_base = nested_array.index(
                    indexes={layer: 0 for layer in layers},
                )

                sizes = {
                    layer: nested_array.sizeof(layers[: i + 1])
                    for i, layer in enumerate(layers)
                }

                full_base_variable_name = f"{variable_base}.{remainder}"
                full_base_variable = f"&{full_base_variable_name}"

            if internal_type == "PackedString":
                meta_entry = []
                variable_base_length_entry = [
                    f".variable_base_length = sizeof({full_base_variable_name}),",
                ]
            else:
                meta_entry = [
                    f".meta_values = {{",
                    meta_initializer,
                    f"}},",
                ]
                variable_base_length_entry = []

            self.common_structure_names[parameter.uuid] = name
            self.h_code.append(
                f"extern InterfaceItem_table_common_{internal_name} {name};",
            )
            self.c_code.extend(
                [
                    f'#pragma DATA_SECTION({name}, "Interface")',
                    f"// {node_path_string(parameter)}",
                    f"// {parameter.uuid}",
                    f"InterfaceItem_table_common_{internal_type} const {name} = {{",
                    [
                        f".common = {{",
                        common_initializers,
                        f"}},",
                        f".variable_base = {full_base_variable},",
                        *variable_base_length_entry,
                        f'.setter = {"NULL" if setter is None else setter},',
                        f'.zone_size = {sizes.get("curve_type", 0)},',
                        f'.curve_size = {sizes.get("curve_index", 0)},',
                        f'.point_size = {sizes.get("point_index", 0)},',
                        *meta_entry,
                    ],
                    f"}};",
                ]
            )

        return name

    def create_item(self, table_element, layers, sunspec_point):
        # TODO: CAMPid 9655426754319431461354643167
        array_element = table_element.original

        if isinstance(array_element, epyqlib.pm.parametermodel.Parameter):
            parameter = array_element
        else:
            parameter = array_element.tree_parent.children[0]

        uses_interface_item = (
            isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
            and parameter.uses_interface_item()
        )

        if not uses_interface_item:
            return [[], []]

        curve_type = get_curve_type("".join(layers[:2]))

        curve_index = int(layers[-2])
        try:
            point_index = int(table_element.name.lstrip("_").lstrip("0")) - 1
        except ValueError:
            point_index = None

        access_level = get_access_level_string(
            parameter=parameter,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        can_signal = self.parameter_uuid_to_can_node.get(table_element.uuid)

        can_getter, can_setter, can_variable = can_getter_setter_variable(
            can_signal,
            parameter,
            var_or_func_or_table="table",
        )

        if can_signal is None:
            can_factor = 1
        else:
            can_factor = can_signal.factor

        # TODO: CAMPid 954679654745154274579654265294624765247569765479
        sunspec_getter = "NULL"
        sunspec_setter = "NULL"
        sunspec_variable = None
        scale_factor_variable = "NULL"
        scale_factor_updater = "NULL"
        sunspec_model_variable = "NULL"

        if sunspec_point is not None:
            sunspec_type = sunspec_types[
                self.parameter_uuid_finder(sunspec_point.type_uuid).name
            ]

            # TODO: CAMPid 9675436715674367943196954756419543975314
            getter_setter_list = [
                "InterfaceItem",
                "table",
                types[parameter.internal_type].name,
                "sunspec",
                sunspec_type,
            ]

            node_in_model = get_sunspec_point_from_table_element(
                sunspec_point=sunspec_point,
                table_element=table_element,
            )

            if node_in_model is not None:
                model_id = node_in_model.tree_parent.tree_parent.id
                sunspec_model_variable = f"sunspecInterface.model{model_id}"
                abbreviation = table_element.abbreviation
                sunspec_variable = (
                    f"{sunspec_model_variable}"
                    f".Curve_{curve_index:>02}_{abbreviation}"
                )

                sunspec_getter = "_".join(
                    str(x) for x in getter_setter_list + ["getter"]
                )
                sunspec_setter = "_".join(
                    str(x) for x in getter_setter_list + ["setter"]
                )

            # TODO: CAMPid 67549654267913467967436
            sunspec_scale_factor = None
            factor_uuid = None
            if node_in_model is not None:
                if node_in_model.factor_uuid is not None:
                    factor_uuid = node_in_model.factor_uuid

            if factor_uuid is not None:
                root = node_in_model.find_root()
                factor_point = root.model.node_from_uuid(
                    node_in_model.factor_uuid,
                )
                sunspec_scale_factor_node = self.parameter_uuid_finder(
                    factor_point.parameter_uuid,
                )
                sunspec_scale_factor = sunspec_scale_factor_node.abbreviation

            if sunspec_scale_factor is not None:
                scale_factor_variable = (
                    f"&{sunspec_model_variable}.{sunspec_scale_factor}"
                )
                scale_factor_updater_name = (
                    f"getSUNSPEC_MODEL{model_id}_{sunspec_scale_factor}"
                )
                scale_factor_updater = f"&{scale_factor_updater_name}"

        common_initializers = create_common_initializers(
            access_level=access_level,
            can_getter=can_getter,
            can_setter=can_setter,
            # not to be used so really hardcode NULL
            can_variable="NULL",
            hand_coded_sunspec_getter_function="NULL",
            hand_coded_sunspec_setter_function="NULL",
            internal_scale=parameter.internal_scale_factor,
            scale_factor_updater=scale_factor_updater,
            scale_factor_variable=scale_factor_variable,
            sunspec_getter=sunspec_getter,
            sunspec_setter=sunspec_setter,
            # not to be used so really hardcode NULL
            sunspec_variable="NULL",
            can_scale_factor=can_factor,
            reject_from_inactive_interfaces=(parameter.reject_from_inactive_interfaces),
            uuid_=table_element.uuid,
            include_uuid_in_item=self.include_uuid_in_item,
        )

        meta_initializer = create_meta_initializer_values(parameter)

        if parameter.internal_variable is None:
            remainder = None
        else:
            remainder = NestedArrays.build(parameter.internal_variable).remainder

        common_structure_name = self.ensure_common_structure(
            internal_type=parameter.internal_type,
            internal_name=types[parameter.internal_type].name,
            parameter=parameter,
            remainder=remainder,
            common_initializers=common_initializers,
            meta_initializer=meta_initializer,
            setter=parameter.setter_function,
        )

        interface_item_type = (
            f"InterfaceItem_table_{types[parameter.internal_type].name}"
        )

        item_uuid_string = str(table_element.uuid).replace("-", "_")
        item_name = f"interfaceItem_{item_uuid_string}"

        maybe_uuid = []
        if self.include_uuid_in_item:
            maybe_uuid = [f".uuid = {uuid_initializer(table_element.uuid)},"]

        maybe_sunspec_variable_length = []
        if sunspec_variable is None:
            sunspec_variable_initializer = "NULL"
        else:
            if parameter.internal_type == "PackedString":
                maybe_sunspec_variable_length = [
                    f".sunspec_variable_length = sizeof({sunspec_variable}),",
                ]
            sunspec_variable_initializer = f"&{sunspec_variable}"

        c = [
            f'#pragma DATA_SECTION({item_name}, "Interface")',
            f"// {node_path_string(table_element)}",
            f"// {table_element.uuid}",
            f"{interface_item_type} const {item_name} = {{",
            [
                f".table_common = &{common_structure_name},",
                f".can_variable = {can_variable},",
                f".sunspec_variable = {sunspec_variable_initializer},",
                *maybe_sunspec_variable_length,
                f'.zone = {curve_type if curve_type is not None else "0"},',
                f".curve = {str(int(curve_index - 1))},",
                f".point = {0 if point_index is None else point_index},",
                *maybe_uuid,
            ],
            "};",
            "",
        ]

        return [
            c,
            [f"extern {interface_item_type} const {item_name};"],
        ]


# TODO: CAMPid 3078980986754174316996743174316967431
def get_sunspec_point_from_table_element(sunspec_point, table_element):
    value = table_element.original

    if isinstance(value, epyqlib.pm.parametermodel.ArrayParameterElement):
        value = value.original

    value = value.uuid

    nodes_in_model = [
        node
        for node in sunspec_point.find_root().nodes_by_attribute(
            attribute_value=value,
            attribute_name="parameter_uuid",
            raise_=False,
        )
        if isinstance(
            node,
            epcpm.sunspecmodel.TableRepeatingBlockReferenceDataPointReference,
        )
    ]

    for node in nodes_in_model:
        for child in node.tree_parent.original.children:
            if child.parameter_uuid == sunspec_point.parameter_uuid:
                node_in_model = node
                break
        else:
            continue

        break
    else:
        node_in_model = None
    return node_in_model


# TODO: CAMPid 3078980986754174316996743174316967431
def get_sunspec_model_from_table_group_element(sunspec_point, table_element):
    nodes_in_model = [
        node
        for node in sunspec_point.find_root().nodes_by_attribute(
            attribute_value=table_element.uuid,
            attribute_name="parameter_uuid",
            raise_=False,
        )
        if isinstance(node, epcpm.sunspecmodel.DataPoint)
    ]

    for node in nodes_in_model:
        for child in node.tree_parent.children:
            if child.parameter_uuid == table_element.uuid:
                node_in_model = node
                break
        else:
            continue

        break
    else:
        return None

    (model_repeating_block,) = [
        node
        for node in sunspec_point.find_root().nodes_by_attribute(
            attribute_value=node_in_model.tree_parent,
            attribute_name="original",
            raise_=False,
        )
        if isinstance(node, epcpm.sunspecmodel.TableRepeatingBlockReference)
    ]

    return model_repeating_block.tree_parent


@builders(epyqlib.pm.parametermodel.Table)
@attr.s
class Table:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        (group,) = (
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.TableGroupElement)
        )

        arrays = [
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.Array)
        ]

        groups = [
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.Group)
        ]

        non_arrays = list(
            itertools.chain.from_iterable(group.children for group in groups)
        )

        # TODO: CAMPid 0795436754762451671643967431
        # TODO: get this from the ...  wherever we have it
        axes = ["x", "y", "z"]

        array_nests = {
            name: NestedArrays.build(s=array.children[0].internal_variable)
            for name, array in zip(axes, arrays)
        }

        non_array_nests = [
            NestedArrays.build(s=non_array.internal_variable)
            for non_array in non_arrays
            if non_array.internal_variable is not None
        ]
        non_array_nests = {nest.remainder: nest for nest in non_array_nests}

        table_base_structures = TableBaseStructures(
            array_nests={**array_nests, **non_array_nests},
            parameter_uuid_to_can_node=self.parameter_uuid_to_can_node,
            parameter_uuid_to_sunspec_node=(self.parameter_uuid_to_sunspec_node),
            parameter_uuid_finder=self.parameter_uuid_finder,
            include_uuid_in_item=self.include_uuid_in_item,
        )

        item_code = builders.wrap(
            wrapped=group,
            can_root=self.can_root,
            sunspec_root=self.sunspec_root,
            table_base_structures=table_base_structures,
            parameter_uuid_to_can_node=self.parameter_uuid_to_can_node,
            parameter_uuid_to_sunspec_node=(self.parameter_uuid_to_sunspec_node),
            parameter_uuid_finder=self.parameter_uuid_finder,
            include_uuid_in_item=self.include_uuid_in_item,
        ).gen()

        return [
            [
                *table_base_structures.c_code,
                "",
                *item_code[0],
            ],
            [
                *table_base_structures.h_code,
                "",
                *item_code[1],
            ],
            set(),
        ]


@builders(epyqlib.pm.parametermodel.TableGroupElement)
@attr.s
class TableGroupElement:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    table_base_structures = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()
    layers = attr.ib(default=[])

    def gen(self):
        c = []
        h = []

        table_tree_root = not isinstance(
            self.wrapped.tree_parent,
            epyqlib.pm.parametermodel.TableGroupElement,
        )

        layers = list(self.layers)
        if not table_tree_root:
            layers.append(self.wrapped.name)

        for child in self.wrapped.children:
            result = builders.wrap(
                wrapped=child,
                can_root=self.can_root,
                sunspec_root=self.sunspec_root,
                table_base_structures=self.table_base_structures,
                parameter_uuid_to_can_node=self.parameter_uuid_to_can_node,
                parameter_uuid_to_sunspec_node=(self.parameter_uuid_to_sunspec_node),
                parameter_uuid_finder=self.parameter_uuid_finder,
                layers=layers,
                include_uuid_in_item=self.include_uuid_in_item,
            ).gen()

            c_built, h_built = result
            c.extend(c_built)
            h.extend(h_built)

        return c, h


# TODO: CAMPid 079549750417808543178043180
def get_curve_type(combination_string):
    # TODO: backmatching
    return {
        "LowRideThrough": "IEEE1547_CURVE_TYPE_LRT",
        "HighRideThrough": "IEEE1547_CURVE_TYPE_HRT",
        "LowTrip": "IEEE1547_CURVE_TYPE_LTRIP",
        "HighTrip": "IEEE1547_CURVE_TYPE_HTRIP",
    }.get(combination_string)


@builders(epyqlib.pm.parametermodel.TableArrayElement)
@attr.s
class TableArrayElement:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    table_base_structures = attr.ib()
    layers = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        table_element = self.wrapped
        zone_node = table_element.tree_parent.tree_parent.tree_parent
        curve_node = zone_node.children[0]
        parameter = curve_node.descendent(
            self.wrapped.tree_parent.name,
            self.wrapped.name,
        )

        sunspec_point = self.parameter_uuid_to_sunspec_node.get(parameter.uuid)

        return self.table_base_structures.create_item(
            table_element=self.wrapped,
            layers=self.layers,
            sunspec_point=sunspec_point,
        )


def create_item(
    item_uuid,
    include_uuid_in_item,
    access_level,
    can_getter,
    can_setter,
    can_variable,
    hand_coded_sunspec_getter_function,
    hand_coded_sunspec_setter_function,
    interface_item_type,
    internal_scale,
    meta_initializer_values,
    parameter,
    scale_factor_updater,
    scale_factor_variable,
    sunspec_getter,
    sunspec_setter,
    sunspec_variable,
    variable_or_getter_setter,
    rejected_callback,
    can_scale_factor,
    reject_from_inactive_interfaces,
):
    item_uuid_string = str(item_uuid).replace("-", "_")
    item_name = f"interfaceItem_{item_uuid_string}"

    if meta_initializer_values is None:
        meta_initializer = []
    else:
        meta_initializer = [
            ".meta_values = {",
            meta_initializer_values,
            "}",
        ]

    common_initializers = create_common_initializers(
        access_level=access_level,
        can_getter=can_getter,
        can_setter=can_setter,
        can_variable=can_variable,
        hand_coded_sunspec_getter_function=hand_coded_sunspec_getter_function,
        hand_coded_sunspec_setter_function=hand_coded_sunspec_setter_function,
        internal_scale=internal_scale,
        scale_factor_updater=scale_factor_updater,
        scale_factor_variable=scale_factor_variable,
        sunspec_getter=sunspec_getter,
        sunspec_setter=sunspec_setter,
        sunspec_variable=sunspec_variable,
        can_scale_factor=can_scale_factor,
        reject_from_inactive_interfaces=reject_from_inactive_interfaces,
        uuid_=item_uuid,
        include_uuid_in_item=include_uuid_in_item,
    )

    item = [
        f'#pragma DATA_SECTION({item_name}, "Interface")',
        f"// {node_path_string(parameter)}",
        f"// {item_uuid}",
        f"{interface_item_type} const {item_name} = {{",
        [
            ".common = {",
            common_initializers,
            "},",
            *variable_or_getter_setter,
            f".rejectedCallback = {rejected_callback},",
            *meta_initializer,
        ],
        "};",
        "",
    ]

    return [
        item,
        [f"extern {interface_item_type} const {item_name};"],
    ]


def uuid_initializer(uuid_):
    return "{{{}}}".format(
        ", ".join(
            "0x{:02x}{:02x}".format(high, low)
            for low, high in toolz.partition_all(2, uuid_.bytes)
        ),
    )


def create_common_initializers(
    access_level,
    can_getter,
    can_setter,
    can_variable,
    hand_coded_sunspec_getter_function,
    hand_coded_sunspec_setter_function,
    internal_scale,
    scale_factor_updater,
    scale_factor_variable,
    sunspec_getter,
    sunspec_setter,
    sunspec_variable,
    can_scale_factor,
    reject_from_inactive_interfaces,
    uuid_,
    include_uuid_in_item,
):
    if can_scale_factor is None:
        # TODO: don't default here?
        can_scale_factor = 1

    maybe_uuid = []
    if include_uuid_in_item:
        maybe_uuid = [f".uuid = {uuid_initializer(uuid_)},"]

    reject_from_inactive_interfaces_literal = (
        "true" if reject_from_inactive_interfaces else "false"
    )

    common_initializers = [
        f".sunspecScaleFactor = {scale_factor_variable},",
        f".canScaleFactor = {float(can_scale_factor)}f,",
        f".scaleFactorUpdater = {scale_factor_updater},",
        f".internalScaleFactor = {internal_scale},",
        f".rejectFromInactiveInterface = {reject_from_inactive_interfaces_literal},",
        f".sunspec = {{",
        [
            f".variable = {sunspec_variable},",
            f".getter = {sunspec_getter},",
            f".setter = {sunspec_setter},",
            f".handGetter = {hand_coded_sunspec_getter_function},",
            f".handSetter = {hand_coded_sunspec_setter_function},",
        ],
        f"}},",
        f".can = {{",
        [
            f".variable = {can_variable},",
            f".getter = {can_getter},",
            f".setter = {can_setter},",
            f".handGetter = NULL,",
            f".handSetter = NULL,",
        ],
        f"}},",
        f".access_level = {access_level},",
        *maybe_uuid,
    ]
    return common_initializers



import os
import pathlib

import attr
import epyqlib.utils.qt
from PyQt5 import QtWidgets

import epcpm.importexportdialog_ui


all_files_filter = ("All Files", ["*"])


def path_or_none(s):
    if isinstance(s, pathlib.Path):
        return s

    if s is None or len(s) == 0:
        return None

    return pathlib.Path(s)


def paths_or_none(x):
    return [pathlib.Path(path) for path in x if len(x) > 0]


def import_dialog():
    dialog = Dialog()

    dialog.ui.sunspec_c_label.hide()
    dialog.ui.sunspec_c.hide()
    dialog.ui.pick_sunspec_c.hide()

    dialog.ui.tables_c_label.hide()
    dialog.ui.tables_c.hide()
    dialog.ui.pick_tables_c.hide()

    dialog.ui.sunspec_tables_c_label.hide()
    dialog.ui.sunspec_tables_c.hide()
    dialog.ui.pick_sunspec_tables_c.hide()

    dialog.ui.sil_tables_c_label.hide()
    dialog.ui.sil_tables_c.hide()
    dialog.ui.pick_sil_tables_c.hide()

    dialog.ui.interface_c_label.hide()
    dialog.ui.interface_c.hide()
    dialog.ui.pick_interface_c.hide()

    dialog.ui.sunspec_bitfields_c_label.hide()
    dialog.ui.sunspec_bitfields_c.hide()
    dialog.ui.pick_sunspec_bitfields_c.hide()

    return dialog


def export_dialog():
    dialog = Dialog(for_save=True)

    dialog.ui.smdx_label.hide()
    dialog.ui.smdx_list.hide()
    dialog.ui.pick_smdx.hide()
    dialog.ui.remove_smdx.hide()
    dialog.setMaximumHeight(0)

    return dialog


@attr.s
class ImportPaths:
    can = attr.ib(converter=path_or_none)
    hierarchy = attr.ib(converter=path_or_none)
    tables_c = attr.ib(converter=path_or_none)
    sunspec_tables_c = attr.ib(converter=path_or_none)
    sunspec_bitfields_c = attr.ib(converter=path_or_none)
    spreadsheet = attr.ib(converter=path_or_none)
    spreadsheet_user = attr.ib(converter=path_or_none)
    smdx = attr.ib(converter=paths_or_none)
    sunspec_c = attr.ib(converter=path_or_none)
    sil_c = attr.ib(converter=path_or_none)
    interface_c = attr.ib(converter=path_or_none)


def paths_from_directory(directory):
    path = pathlib.Path(directory)
    interface = path / "interface"
    embedded = path / "embedded-library"
    sunspec = embedded / "system" / "sunspec"

    return ImportPaths(
        can=interface / "EPC_DG_ID247_FACTORY.sym",
        hierarchy=interface / "EPC_DG_ID247_FACTORY.parameters.json",
        tables_c=interface / "canInterfaceGenTables.c",
        sunspec_tables_c=sunspec / "sunspecInterfaceGenTables.c",
        sunspec_bitfields_c=sunspec / "sunspecInterfaceBitfieldsGen.c",
        spreadsheet=embedded / "MODBUS_SunSpec-EPC.xlsx",
        spreadsheet_user=embedded / "EPCSunspec.xlsx",
        smdx=sorted(sunspec.glob("smdx_*.xml")),
        sunspec_c=sunspec,
        sil_c=path / "sil" / "libEpcControlInterfaceGen.c",
        interface_c=interface / "interfaceGen.c",
    )


@attr.s
class Dialog(QtWidgets.QDialog):
    ui = attr.ib(factory=epcpm.importexportdialog_ui.Ui_Dialog)
    paths_result = attr.ib(default=None)
    for_save = attr.ib(default=False)
    _parent = attr.ib(default=None)
    directory = attr.ib(default=None)

    def __attrs_post_init__(self):
        super().__init__(self._parent)

        self.ui.setupUi(self)

        self.ui.buttons.accepted.connect(self.accept)
        self.ui.buttons.rejected.connect(self.reject)

        self.ui.pick_can.clicked.connect(self.pick_can)
        self.ui.pick_hierarchy.clicked.connect(self.pick_hierarchy)
        self.ui.pick_tables_c.clicked.connect(self.pick_tables_c)
        self.ui.pick_sunspec_tables_c.clicked.connect(
            self.pick_sunspec_tables_c,
        )
        self.ui.pick_spreadsheet.clicked.connect(self.pick_spreadsheet)
        self.ui.pick_spreadsheet_user.clicked.connect(self.pick_spreadsheet_user)
        self.ui.pick_sunspec_c.clicked.connect(self.pick_sunspec_c)
        self.ui.pick_sunspec_bitfields_c.clicked.connect(self.pick_sunspec_bitfields_c)
        self.ui.pick_sil_c.clicked.connect(self.pick_sil_c)
        self.ui.pick_interface_c.clicked.connect(self.pick_interface_c)
        self.ui.pick_smdx.clicked.connect(self.pick_smdx)
        self.ui.remove_smdx.clicked.connect(self.remove_smdx)
        self.ui.from_directory.clicked.connect(self.from_directory)

    def accept(self):
        if self.for_save:
            smdx = []
        else:
            smdx = self.smdx_paths()

        self.paths_result = ImportPaths(
            can=self.ui.can.text(),
            hierarchy=self.ui.hierarchy.text(),
            tables_c=self.ui.tables_c.text(),
            sunspec_tables_c=self.ui.sunspec_tables_c.text(),
            sunspec_bitfields_c=self.ui.sunspec_bitfields_c.text(),
            spreadsheet=self.ui.spreadsheet.text(),
            spreadsheet_user=self.ui.spreadsheet_user.text(),
            smdx=smdx,
            sunspec_c=self.ui.sunspec_c.text(),
            sil_c=self.ui.sil_c.text(),
            interface_c=self.ui.interface_c.text(),
        )
        super().accept()

    def from_directory(self):
        directory = QtWidgets.QFileDialog.getExistingDirectory(parent=self)
        if len(directory) == 0:
            return

        self.directory = pathlib.Path(directory)

        paths = paths_from_directory(directory=self.directory)

        self.ui.can.setText(os.fspath(paths.can))
        self.ui.hierarchy.setText(os.fspath(paths.hierarchy))
        self.ui.tables_c.setText(os.fspath(paths.tables_c))
        self.ui.sunspec_tables_c.setText(os.fspath(paths.sunspec_tables_c))
        self.ui.sunspec_bitfields_c.setText(os.fspath(paths.sunspec_bitfields_c))
        self.ui.spreadsheet.setText(os.fspath(paths.spreadsheet))
        self.ui.spreadsheet_user.setText(os.fspath(paths.spreadsheet_user))
        self.ui.sunspec_c.setText(os.fspath(paths.sunspec_c))
        self.ui.sil_c.setText(os.fspath(paths.sil_c))
        self.ui.interface_c.setText(os.fspath(paths.interface_c))

        self.clear_smdx_list()
        if not self.for_save:
            for path in paths.smdx:
                self.ui.smdx_list.addItem(os.fspath(path))

    def clear_smdx_list(self):
        for i in range(self.ui.smdx_list.count()):
            self.ui.smdx_list.takeAt(i)

    def smdx_paths(self):
        return [
            self.ui.smdx_list.item(index).text()
            for index in range(self.ui.smdx_list.count())
        ]

    def remove_smdx(self):
        selected_items = self.ui.smdx_list.selectedItems()

        for item in reversed(selected_items):
            self.ui.smdx_list.takeItem(self.ui.smdx_list.row(item))

    def pick_can(self):
        filters = (
            ("PEAK PCAN Symbol File", ["sym"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.can,
            filters=filters,
            multiple=False,
        )

    def pick_hierarchy(self):
        filters = (
            ("Parameter Hierarchy", ["json"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.hierarchy,
            filters=filters,
            multiple=False,
        )

    def pick_tables_c(self):
        filters = (
            ("Tables C", ["c"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.tables_c,
            filters=filters,
            multiple=False,
        )

    def pick_sunspec_tables_c(self):
        filters = (
            ("SunSpec Tables C", ["c"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.sunspec_tables_c,
            filters=filters,
            multiple=False,
        )

    def pick_sunspec_bitfields_c(self):
        filters = (
            ("SunSpec Bitfields C", ["c"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.sunspec_bitfields_c,
            filters=filters,
            multiple=False,
        )

    def pick_spreadsheet(self):
        filters = (
            ("SunSpec Spreadsheet", ["xls"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.spreadsheet,
            filters=filters,
            multiple=False,
        )

    def pick_spreadsheet_user(self):
        filters = (
            ("SunSpec Spreadsheet", ["xls"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.spreadsheet_user,
            filters=filters,
            multiple=False,
        )

    def pick_sunspec_c(self):
        directory = QtWidgets.QFileDialog.getExistingDirectory(parent=self)

        if len(directory) == 0:
            return

        self.ui.sunspec_c.setText(directory)

    def pick_sil_c(self):
        directory = QtWidgets.QFileDialog.getExistingDirectory(parent=self)

        if len(directory) == 0:
            return

        self.ui.sil_c.setText(directory)

    def pick_interface_c(self):
        filters = (
            ("Tables C", ["c"]),
            all_files_filter,
        )

        self.file_dialog(
            target=self.ui.sunspec_tables_c,
            filters=filters,
            multiple=False,
        )

    def pick_smdx(self):
        filters = (
            ("SunSpec SMDX", ["xml"]),
            all_files_filter,
        )

        paths = self.file_dialog(
            target=None,
            filters=filters,
            multiple=True,
        )

        existing_items = set(self.smdx_paths())

        for path in paths:
            path = os.fspath(path)

            if path not in existing_items:
                self.ui.smdx_list.addItem(path)

    def file_dialog(self, target, filters, multiple):
        path = epyqlib.utils.qt.file_dialog(
            filters=filters,
            parent=self,
            save=self.for_save,
            path_factory=pathlib.Path,
            multiple=multiple,
        )

        if path is None:
            return

        if target is not None:
            target.setText(os.fspath(path))

        return path


@attr.s
class Main:
    application = attr.ib()
    dialog = attr.ib(factory=Dialog)

    def show_dialog(self):
        self.dialog.accepted.connect(self.dialog_accepted)
        self.dialog.open()

    def dialog_accepted(self):
        print(self.dialog.paths_result)
        self.application.quit()

    def one_call(self):
        if self.dialog.exec():
            print(self.dialog.paths_result)
        self.application.quit()


def main():
    from PyQt5 import QtCore
    from PyQt5 import QtWidgets

    app = QtWidgets.QApplication([])

    main = Main(application=app)
    QtCore.QTimer.singleShot(0, main.one_call)
    app.exec()


if __name__ == "__main__":
    main()



"""
Testing for parameterstoc script to build a parameters model.

It is possible that the parameterstoc module is either out of date
or no longer applicable for normal use.
"""
import functools
import os
import textwrap

import click.testing
import pycparser.c_ast
import pycparser.c_generator
import pytest

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel

import epcpm.parameterstoc


def disabled_test_exploration():
    path = os.path.join(os.path.dirname(__file__), "example_parameters.json")

    runner = click.testing.CliRunner()
    runner.isolated_filesystem()
    result = runner.invoke(
        epcpm.parameterstoc.cli,
        [
            "--parameters",
            path,
        ],
    )

    print(result.output)

    assert result.exit_code == 0


def test_pycparser_exploration_parse():
    sample = """
    typedef int int16_t;
    typedef int uint16_t;

    enum EnumName
    {
        one=1,
        two=2
    };
    typedef enum EnumName EnumTypedefName;

    struct StructName
    {
        int16_t a;
        uint16_t b;
    };
    typedef struct StructName StructTypedefName;
    
    int16_t array[5];
    """

    parser = pycparser.CParser()
    ast = parser.parse(sample)

    generator = pycparser.c_generator.CGenerator()

    generator.visit(ast)

    return ast


def test_pycparser_exploration_wrapped():
    top_level = []

    top_level.extend(
        epcpm.parameterstoc.enum(
            name="EnumName",
            enumerators=(
                ("a", 1),
                ("b", 2),
            ),
        )
    )

    top_level.extend(
        epcpm.parameterstoc.struct(
            name="StructName",
            member_decls=(
                epcpm.parameterstoc.Decl(
                    type=epcpm.parameterstoc.Type(
                        name=name,
                        type=type,
                    )
                )
                for type, name in (
                    ("int16_t", "a"),
                    ("uint16_t", "b"),
                )
            ),
        )
    )

    ast = pycparser.c_ast.FileAST(top_level)

    generator = pycparser.c_generator.CGenerator()

    s = generator.visit(ast)
    assert s == textwrap.dedent(
        """\
    enum EnumName_e
    {
      a = 1,
      b = 2
    };
    typedef enum EnumName_e EnumName_et;
    struct StructName_s
    {
      int16_t a;
      uint16_t b;
    };
    typedef struct StructName_s
    {
    } StructName_t;
    """
    )


def test_single_layer_group_to_c():
    group = epyqlib.pm.parametermodel.Group(
        name="Group Name",
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter A",
            type_name="int16_t",
        ),
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter B",
            type_name="int16_t",
        ),
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter C",
            type_name="int16_t",
        ),
    ]

    for child in children:
        group.append_child(child)

    builder = epcpm.parameterstoc.builders.wrap(group)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
        struct GroupName_s
        {
          int16_t parameterA;
          int16_t parameterB;
          int16_t parameterC;
        };
        typedef struct GroupName_s
        {
        } GroupName_t;
        """
    )


def test_nested_group_to_c():
    inner_inner_group = epyqlib.pm.parametermodel.Group(
        name="Inner Inner Group Name",
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter F",
            type_name="int16_t",
        ),
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter G",
            type_name="int16_t",
        ),
    ]

    for child in children:
        inner_inner_group.append_child(child)

    inner_group = epyqlib.pm.parametermodel.Group(
        name="Inner Group Name",
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter D",
            type_name="int16_t",
        ),
        inner_inner_group,
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter E",
            type_name="int16_t",
        ),
    ]

    for child in children:
        inner_group.append_child(child)

    outer_group = epyqlib.pm.parametermodel.Group(
        name="Outer Group Name",
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter A",
            type_name="int16_t",
        ),
        inner_group,
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter B",
            type_name="int16_t",
        ),
        epyqlib.pm.parametermodel.Parameter(
            name="Parameter C",
            type_name="int16_t",
        ),
    ]

    for child in children:
        outer_group.append_child(child)

    builder = epcpm.parameterstoc.builders.wrap(outer_group)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
        struct InnerInnerGroupName_s
        {
          int16_t parameterF;
          int16_t parameterG;
        };
        typedef struct InnerInnerGroupName_s
        {
        } InnerInnerGroupName_t;
        struct InnerGroupName_s
        {
          int16_t parameterD;
          InnerInnerGroupName_t innerInnerGroupName;
          int16_t parameterE;
        };
        typedef struct InnerGroupName_s
        {
        } InnerGroupName_t;
        struct OuterGroupName_s
        {
          int16_t parameterA;
          InnerGroupName_t innerGroupName;
          int16_t parameterB;
          int16_t parameterC;
        };
        typedef struct OuterGroupName_s
        {
        } OuterGroupName_t;
        """
    )


def test_datalogger_a():
    data_logger = epyqlib.pm.parametermodel.Group(name="Data Logger")

    chunks_array = epyqlib.pm.parametermodel.Array(name="Chunks")
    data_logger.append_child(chunks_array)

    chunk = epyqlib.pm.parametermodel.Group(type_name="Chunk")
    chunks_array.append_child(chunk)
    chunks_array.length = 4
    chunks_array.children[0].name = "First"
    chunks_array.children[1].name = "Second"
    chunks_array.children[2].name = "Third"
    chunks_array.children[3].name = "Fourth"

    address = epyqlib.pm.parametermodel.Parameter(
        name="Address",
        default=0,
        type_name="int16_t",
    )
    chunk.append_child(address)
    bytes_ = epyqlib.pm.parametermodel.Parameter(
        name="Bytes",
        default=0,
        type_name="int16_t",
    )
    chunk.append_child(bytes_)

    post_trigger_duration = epyqlib.pm.parametermodel.Parameter(
        name="Post Trigger Duration",
        default=500,
        type_name="int16_t",
    )
    data_logger.append_child(post_trigger_duration)

    group = epyqlib.pm.parametermodel.Group(name="Group")
    data_logger.append_child(group)

    param = epyqlib.pm.parametermodel.Parameter(
        name="Param",
        type_name="int16_t",
    )
    group.append_child(param)

    builder = epcpm.parameterstoc.builders.wrap(data_logger)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    struct Chunk_s
    {
      int16_t address;
      int16_t bytes;
    };
    typedef struct Chunk_s
    {
    } Chunk_t;
    enum Chunks_e
    {
      Chunks_First = 0,
      Chunks_Second = 1,
      Chunks_Third = 2,
      Chunks_Fourth = 3,
      Chunks_Count = 4
    };
    typedef enum Chunks_e Chunks_et;
    typedef Chunk_t Chunks_t[Chunks_Count];
    struct Group_s
    {
      int16_t param;
    };
    typedef struct Group_s
    {
    } Group_t;
    struct DataLogger_s
    {
      Chunks_t chunks;
      int16_t postTriggerDuration;
      Group_t group;
    };
    typedef struct DataLogger_s
    {
    } DataLogger_t;
    """
    )


data_logger_structures = """\
typedef struct
{
    void*	address;
    size_t	bytes;
} DataLogger_Chunk;

#define DATALOGGER_CHUNK_DEFAULTS(...) \
{ \
    .address = 0, \
    .bytes = 0, \
    __VA_ARGS__ \
}

typedef DataLogger_Chunk DataLogger_Chunks[dataLoggerChunkCount];

typedef struct
{
    DataLogger_Chunks chunks;
    uint16_t postTriggerDuration_ms;
} DataLogger_Params;

#define DATALOGGER_PARAMS_DEFAULTS(...) \
{ \
    .chunks = {[0 ... dataLoggerChunkCount-1] = DATALOGGER_CHUNK_DEFAULTS()}, \
    .postTriggerDuration_ms = 500, \
    __VA_ARGS__ \
}
"""


def test_basic_parameter_array():
    array = epyqlib.pm.parametermodel.Array(name="Array Name")
    parameter = epyqlib.pm.parametermodel.Parameter(
        name="Parameter Name",
        type_name="int16_t",
    )

    array.append_child(parameter)

    n = 5

    array.length = n

    builder = epcpm.parameterstoc.builders.wrap(array)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    enum ArrayName_e
    {
      ArrayName_ParameterName = 0,
      ArrayName__2 = 1,
      ArrayName__3 = 2,
      ArrayName__4 = 3,
      ArrayName__5 = 4,
      ArrayName_Count = 5
    };
    typedef enum ArrayName_e ArrayName_et;
    typedef int16_t ArrayName_t[ArrayName_Count];
    """
    )


def test_grouped_parameter_array():
    group = epyqlib.pm.parametermodel.Group(name="Group Name")
    array = epyqlib.pm.parametermodel.Array(name="Array Name")
    parameter = epyqlib.pm.parametermodel.Parameter(
        name="Parameter Name",
        type_name="int16_t",
    )

    group.append_child(array)
    array.append_child(parameter)

    n = 5

    array.length = n

    builder = epcpm.parameterstoc.builders.wrap(group)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    enum ArrayName_e
    {
      ArrayName_ParameterName = 0,
      ArrayName__2 = 1,
      ArrayName__3 = 2,
      ArrayName__4 = 3,
      ArrayName__5 = 4,
      ArrayName_Count = 5
    };
    typedef enum ArrayName_e ArrayName_et;
    typedef int16_t ArrayName_t[ArrayName_Count];
    struct GroupName_s
    {
      ArrayName_t arrayName;
    };
    typedef struct GroupName_s
    {
    } GroupName_t;
    """
    )


def test_grouped_parameter_array_no_enum():
    array = epyqlib.pm.parametermodel.Array(name="Array Name")
    array.named_enumerators = False
    parameter = epyqlib.pm.parametermodel.Parameter(
        name="Parameter Name",
        type_name="int16_t",
    )

    array.append_child(parameter)

    n = 5

    array.length = n

    builder = epcpm.parameterstoc.builders.wrap(array)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    enum ArrayName_e
    {
      ArrayName_Count = 5
    };
    typedef enum ArrayName_e ArrayName_et;
    typedef int16_t ArrayName_t[ArrayName_Count];
    """
    )


def test_line_monitor_params():
    line_monitoring = epyqlib.pm.parametermodel.Group(
        name="Line Monitoring",
    )

    frequency_limits = epyqlib.pm.parametermodel.Array(
        name="Frequency Limits",
    )
    line_monitoring.append_child(frequency_limits)

    frequency_limit = epyqlib.pm.parametermodel.Group(
        name="First",
        type_name="Frequency Limit",
    )
    frequency_limits.append_child(frequency_limit)

    frequency = epyqlib.pm.parametermodel.Parameter(
        name="Frequency",
        type_name="_iq",
    )
    frequency_limit.append_child(frequency)

    clearing_time = epyqlib.pm.parametermodel.Parameter(
        name="Clearing Time",
        type_name="_iq",
    )
    frequency_limit.append_child(clearing_time)

    frequency_limits.length = 4
    frequency_limits.children[1].name = "Second"
    frequency_limits.children[2].name = "Third"
    frequency_limits.children[3].name = "Fourth"

    builder = epcpm.parameterstoc.builders.wrap(line_monitoring)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    struct FrequencyLimit_s
    {
      _iq frequency;
      _iq clearingTime;
    };
    typedef struct FrequencyLimit_s
    {
    } FrequencyLimit_t;
    enum FrequencyLimits_e
    {
      FrequencyLimits_First = 0,
      FrequencyLimits_Second = 1,
      FrequencyLimits_Third = 2,
      FrequencyLimits_Fourth = 3,
      FrequencyLimits_Count = 4
    };
    typedef enum FrequencyLimits_e FrequencyLimits_et;
    typedef FrequencyLimit_t FrequencyLimits_t[FrequencyLimits_Count];
    struct LineMonitoring_s
    {
      FrequencyLimits_t frequencyLimits;
    };
    typedef struct LineMonitoring_s
    {
    } LineMonitoring_t;
    """
    )


def test_root():
    root = epyqlib.pm.parametermodel.Root()
    group = epyqlib.pm.parametermodel.Group(name="Group")
    p1 = epyqlib.pm.parametermodel.Parameter(name="red", type_name="RedType")
    p2 = epyqlib.pm.parametermodel.Parameter(name="blue", type_name="BlueType")

    root.append_child(group)
    group.append_child(p1)
    root.append_child(p2)

    builder = epcpm.parameterstoc.builders.wrap(root)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    struct Group_s
    {
      RedType red;
    };
    typedef struct Group_s
    {
    } Group_t;
    """
    )

    ast = pycparser.c_ast.FileAST(builder.instantiation())
    s = generator.visit(ast)

    assert s == textwrap.dedent(
        """\
    Group_t group;
    BlueType blue;
    """
    )



import collections
import contextlib
import os

import attr
import epyqlib.pm.parametermodel
import openpyxl
import sunspec.core.device

import epcpm.sunspecmodel


def epc_point_from_pysunspec_point(
    point, parameter_model, parameter_uuid, scale_factors=None
):
    if scale_factors is not None and point.point_type.sf is not None:
        scale_factor_uuid = scale_factors[point.point_type.sf].uuid
    else:
        scale_factor_uuid = None

    sunspec_type_uuid = point.point_type.type
    if sunspec_type_uuid is not None:
        root = parameter_model.list_selection_roots["sunspec types"]
        sunspec_type_uuid = root.child_by_name(sunspec_type_uuid).uuid

    return epcpm.sunspecmodel.DataPoint(
        factor_uuid=scale_factor_uuid,
        parameter_uuid=parameter_uuid,
        type_uuid=sunspec_type_uuid,
        # enumeration_uuid=,
        block_offset=point.point_type.offset,
        size=point.point_type.len,
        mandatory=point.point_type.mandatory == "true",
        # uuid=,
    )


def epc_parameter_from_pysunspec_point(point):
    parameter = epyqlib.pm.parametermodel.Parameter(
        name=point.point_type.label,
        abbreviation=point.point_type.id,
        notes=point.point_type.notes,
        units=point.point_type.units,
        comment=point.point_type.description,
        read_only="w" not in point.point_type.access,
    )

    return parameter


def import_models(*model_ids, parameter_model, paths):
    return [
        import_model(model_id=id, parameter_model=parameter_model, paths=paths)
        for id in model_ids
    ]


@contextlib.contextmanager
def fresh_smdx_path(*paths):
    original_pathlist = sunspec.core.device.file_pathlist
    sunspec.core.device.file_pathlist = sunspec.core.util.PathList()

    for path in paths:
        sunspec.core.device.file_pathlist.add(os.fspath(path))

    try:
        yield sunspec.core.device.file_pathlist
    finally:
        sunspec.core.device.file_pathlist = original_pathlist


def none_to_empty_string(value):
    if value is None:
        return ""

    return value


@attr.s(frozen=True)
class Symbol:
    value = attr.ib(converter=int)
    id = attr.ib()
    label = attr.ib()
    type = attr.ib()
    description = attr.ib(converter=none_to_empty_string)
    notes = attr.ib(converter=none_to_empty_string)

    @classmethod
    def from_sunspec(cls, symbol, type_):
        return cls(
            description=symbol.description,
            id=symbol.id,
            label=symbol.label,
            notes=symbol.notes,
            value=symbol.value,
            type=type_,
        )


def import_model(model_id, parameter_model, paths=()):
    model = sunspec.core.device.Model(mid=model_id)
    if len(paths) == 0:
        model.load()
    else:
        with fresh_smdx_path(*paths):
            model.load()

    imported_points = []
    scale_factors = {}

    group = epyqlib.pm.parametermodel.Group(
        name="SunSpec Model {}".format(model.id),
    )
    parameter_model.root.append_child(group)

    our_model = epcpm.sunspecmodel.Model(
        id=model.id,
        length=model.len,
    )

    types = parameter_model.list_selection_roots["sunspec types"]
    parameters = our_model.children[0].add_data_points(
        model_id=model.model_type.label,
        uint16_uuid=types.child_by_name("uint16").uuid,
    )

    for parameter in parameters:
        group.append_child(parameter)

    for name, point in model.points_sf.items():
        parameter = epc_parameter_from_pysunspec_point(point=point)
        group.append_child(parameter)
        epc_point = epc_point_from_pysunspec_point(
            point=point,
            parameter_uuid=parameter.uuid,
            parameter_model=parameter_model,
        )
        scale_factors[name] = epc_point

    enumerations = collections.defaultdict(list)

    for point in model.points_list:
        parameter = epc_parameter_from_pysunspec_point(point=point)
        group.append_child(parameter)
        epc_point = epc_point_from_pysunspec_point(
            point=point,
            parameter_model=parameter_model,
            parameter_uuid=parameter.uuid,
            scale_factors=scale_factors,
        )

        imported_points.append(epc_point)

        if point.point_type.type.startswith(("enum", "bitfield")):
            enumeration = tuple(
                sorted(
                    Symbol.from_sunspec(symbol=symbol, type_=point.point_type.type)
                    for symbol in point.point_type.symbols
                )
            )
            enumerations[enumeration].append(epc_point)

    enumerations_root = parameter_model.list_selection_roots["enumerations"]
    for enumeration, points in enumerations.items():
        parameter = parameter_model.node_from_uuid(points[0].parameter_uuid)
        # TODO: just using the first point?  hmm
        epc_enumeration = epyqlib.pm.parametermodel.Enumeration(
            name="SunSpec{}".format(parameter.name),
        )

        points[0].enumeration_uuid = epc_enumeration.uuid

        for symbol in enumeration:
            enumerator = epyqlib.pm.parametermodel.SunSpecEnumerator(
                name=symbol.id,
                label=symbol.label,
                description=symbol.description,
                notes=symbol.notes,
                value=symbol.value,
                type=symbol.type,
            )
            epc_enumeration.append_child(enumerator)

        enumerations_root.append_child(epc_enumeration)

    id_point = our_model.children[0].children[0]
    id_point.id = model.model_type.id
    id_point.notes = model.model_type.notes

    parameters[0].comment = model.model_type.description

    imported_points = sorted(
        imported_points + list(scale_factors.values()),
        key=lambda point: point.block_offset,
    )

    for point in imported_points:
        our_model.children[1].append_child(point)

    return our_model


@attr.s(frozen=True)
class GetSetKey:
    model = attr.ib()
    name = attr.ib()
    get_set = attr.ib()


def import_get_set(path):
    workbook = openpyxl.load_workbook(path)

    collected = {}

    for sheet in workbook.worksheets:
        try:
            model = int(sheet.title)
        except ValueError:
            continue

        iter_rows = iter(sheet.rows)

        column_indexes = {cell.value: i for i, cell in enumerate(next(iter_rows))}

        for row in iter_rows:
            for get_set in ("get", "set"):
                value = row[column_indexes[get_set]].value

                if value in (None, ""):
                    continue

                name = row[column_indexes["Name"]].value
                key = GetSetKey(
                    model=model,
                    name=name,
                    get_set=get_set,
                )
                collected[key] = value

    return collected



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['epcpm', 'epcpm.cli', 'epcpm.tests', 'epcpm.tests.cli']

package_data = \
{'': ['*'], 'epcpm.tests': ['project/*', 'sunspec/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'PyQt5==5.13.0',
 'Twisted==21.2.0',
 'attrs==20.2.0',
 'canmatrix==0.9.1',
 'click>=7.0.0,<8.0.0',
 'codecov>=2.1.12,<3.0.0',
 'graham==0.1.11',
 'lxml==4.3.0',
 'openpyxl==2.5.12',
 'pycparser==2.21',
 'pyserial==3.4',
 'pysunspec>=2.1.1,<3.0.0',
 'pytest-cov==2.5.1',
 'pytest-qt==3.2.1',
 'pytest-twisted==1.9',
 'pytest==3.8.2',
 'python-can>=3.3.4,<4.0.0',
 'requests==2.26.0',
 'sunspecdemo>=0.1.9,<0.2.0',
 'toolz==0.9.0',
 'tox==2.8.2',
 'twine==1.13.0',
 'xmldiff==2.2']

entry_points = \
{'console_scripts': ['buildui = buildui:compile_ui',
                     'epcconvertparameters = epcpm.cli.convertepp:cli',
                     'epcexportdocx = epcpm.cli.exportdocx:cli',
                     'epcexportsym = epcpm.cli.exportsym:cli',
                     'epcimportsym = epcpm.cli.importsym:cli',
                     'epcparameterstoc = epcpm.cli.parameterstoc:cli',
                     'epcpm = epcpm.__main__:_entry_point',
                     'epcpmcli = epcpm.cli.main:main']}

setup_kwargs = {
    'name': 'epcpm',
    'version': '2022.1.7',
    'description': '',
    'long_description': None,
    'author': 'Alex Anker',
    'author_email': 'alex.anker@epcpower.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import itertools

import attr
import graham
import marshmallow

import epyqlib.attrsmodel
import epyqlib.checkresultmodel
import epyqlib.pm.parametermodel
import epyqlib.utils
import epyqlib.utils.qt
from PyQt5 import QtCore
from PyQt5 import QtWidgets


class ConsistencyError(Exception):
    pass


class MismatchedSizeAndTypeError(Exception):
    pass


class TypeNotFoundError(Exception):
    pass


def build_sunspec_types_enumeration():
    enumeration = epyqlib.pm.parametermodel.Enumeration(
        name="SunSpecTypes",
        uuid="00b90651-3e3b-4e28-a8c0-7339ae092200",
    )

    enumerators = [
        epyqlib.pm.parametermodel.Enumerator(
            name="int16", value=1, uuid="2cf75e5a-ffc8-422a-bbc6-573d4206a7e1"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="uint16", value=1, uuid="4f856a7e-20f4-43e2-86b1-cc7ee772f919"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="int32", value=2, uuid="4fec39a5-b702-4dbf-8ad1-95f5e01201b6"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="uint32", value=2, uuid="eb8cdc87-05e2-4593-994e-ab3363236168"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="sunssf", value=1, uuid="02e70616-4986-4f3e-8ac4-98ac153e66f9"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="enum16", value=1, uuid="209aebc8-652f-47bf-9952-4c112ced2781"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="bitfield16", value=1, uuid="5d30a559-13fe-42b1-89df-35c3edc237f0"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="bitfield32", value=2, uuid="fc0ad957-2785-4762-b2fc-4db2cf785ca2"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="string", value=0, uuid="5460c860-4aad-476a-908c-83a364b781c9"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="acc16", value=1, uuid="05830309-c61c-41d4-8c66-88ed25187575"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="acc32", value=2, uuid="f9d30fa6-33b2-48a2-8b64-72a4f47c0bd4"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="acc64", value=2, uuid="7ba30a78-a76d-4ce7-bb32-08e9410b14b9"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="count", value=2, uuid="d5d00207-d2c4-413f-88eb-e44946350007"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="pad", value=1, uuid="f8090bab-cf12-476c-b96a-1c8bb9848bb5"
        ),
    ]

    for enumerator in enumerators:
        enumeration.append_child(enumerator)

    return enumeration


def create_size_attribute(default=0):
    return attr.ib(
        default=default,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )


def create_factor_uuid_attribute():
    return epyqlib.attrsmodel.attr_uuid(
        default=None,
        human_name="Scale Factor",
        allow_none=True,
        data_display=name_from_uuid,
        list_selection_path=("..", "..", "Fixed Block"),
        override_delegate=ScaleFactorDelegate,
    )


def create_parameter_uuid_attribute():
    return epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Parameter",
        data_display=name_from_uuid_and_parent,
        editable=False,
    )


# TODO: CAMPid 8695426542167924656654271657917491654
def name_from_uuid(node, value, model):
    if value is None:
        return None

    try:
        target_node = model.node_from_uuid(value)
    except epyqlib.attrsmodel.NotFoundError:
        return str(value)

    return model.node_from_uuid(target_node.parameter_uuid).abbreviation


# TODO: CAMPid 8695426542167924656654271657917491654
def name_from_uuid_and_parent(node, value, model):
    if value is None:
        return None

    try:
        target_node = model.node_from_uuid(value)
    except epyqlib.attrsmodel.NotFoundError:
        return str(value)

    return "{} - {}".format(target_node.tree_parent.name, target_node.name)


class ScaleFactorDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, text_column_name, root, parent):
        super().__init__(parent)

        self.root = root

    def createEditor(self, parent, option, index):
        return QtWidgets.QListWidget(parent=parent)

    def setEditorData(self, editor, index):
        model_index = epyqlib.attrsmodel.to_source_model(index)
        model = model_index.model()

        item = model.itemFromIndex(model_index)
        attrs_model = item.data(epyqlib.utils.qt.UserRoles.attrs_model)

        raw = model.data(model_index, epyqlib.utils.qt.UserRoles.raw)

        points = []
        for pt in self.root.children:
            type_node = attrs_model.node_from_uuid(pt.type_uuid)
            if type_node.name == "sunssf":
                points.append(pt)

        it = QtWidgets.QListWidgetItem(editor)
        it.setText("")
        it.setData(epyqlib.utils.qt.UserRoles.raw, "")
        it.setSelected(True)
        for p in points:
            it = QtWidgets.QListWidgetItem(editor)
            param = attrs_model.node_from_uuid(p.parameter_uuid)
            it.setText(param.abbreviation)
            it.setData(epyqlib.utils.qt.UserRoles.raw, p.uuid)
            if p.uuid == raw:
                it.setSelected(True)

        editor.setMinimumHeight(editor.sizeHint().height())
        editor.itemClicked.connect(
            lambda: epyqlib.attrsmodel.hide_popup(editor),
        )
        editor.show()

    def setModelData(self, editor, model, index):
        selected_item = editor.currentItem()
        datum = str(selected_item.data(epyqlib.utils.qt.UserRoles.raw))
        model.setData(index, datum)


@graham.schemify(tag="data_point", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class DataPoint(epyqlib.treenode.TreeNode):
    factor_uuid = create_factor_uuid_attribute()
    parameter_uuid = create_parameter_uuid_attribute()

    hand_coded_getter = epyqlib.attrsmodel.create_checkbox_attribute(
        default=False,
    )
    hand_coded_setter = epyqlib.attrsmodel.create_checkbox_attribute(
        default=False,
    )

    not_implemented = epyqlib.attrsmodel.create_checkbox_attribute(
        default=False,
    )

    type_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Type",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root="sunspec types",
    )

    offset = attr.ib(
        default=0,
        converter=int,
    )  # this is somewhat redundant with the position in the list of data
    # points but helpful for keeping things from incidentally floating
    # around especially in custom models where we have no sunspec
    # model to be validating against
    # for now, yes, this is vaguely nondescript of address vs block offset
    @QtCore.pyqtProperty("PyQt_PyObject")
    def pyqtify_offset(self):
        block_offset = getattr(self.tree_parent, "block_offset", None)

        if block_offset is None:
            return None

        return block_offset + self.block_offset

    # TODO: shouldn't this be read only?
    @pyqtify_offset.setter
    def pyqtify_offset(self, value):
        pass

    block_offset = attr.ib(
        default=0,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )

    size = create_size_attribute()

    enumeration_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=enumeration_uuid,
        human_name="Enumeration",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root="enumerations",
        ),
    )

    units = epyqlib.attrsmodel.create_str_or_none_attribute()

    get = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=get,
        no_column=True,
    )
    set = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_str_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(allow_none=True),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=set,
        no_column=True,
    )

    mandatory = attr.ib(
        default=True,
        converter=epyqlib.attrsmodel.two_state_checkbox,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Boolean(),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()
    # value  doesn't make sense here, this is interface definition, not a
    #       value set

    # ? point_id = attr.ib()
    # ? index = index

    # last access time?
    # time = time

    #     getter_code = attr.ib()
    #     setter_code = attr.ib()
    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, epyqlib.pm.parametermodel.Parameter)

    def child_from(self, node):
        self.parameter_uuid = node.uuid

        return None

    @epyqlib.attrsmodel.check_children
    def check(self, result, models):
        if self.parameter_uuid is None:
            result.append_child(
                epyqlib.checkresultmodel.Result(
                    node=self,
                    severity=epyqlib.checkresultmodel.ResultSeverity.error,
                    message="No parameter connected",
                )
            )
        else:
            root = self.find_root()
            parameter = root.model.node_from_uuid(self.parameter_uuid)

            if (
                parameter.tree_parent.name.startswith("SunSpec")
                and parameter.tree_parent.tree_parent.tree_parent is None
            ):
                result.append_child(
                    epyqlib.checkresultmodel.Result(
                        node=self,
                        severity=(epyqlib.checkresultmodel.ResultSeverity.information),
                        message=("Connected to temporary SunSpec imported parameter"),
                    )
                )

            if not parameter.uses_interface_item():
                result.append_child(
                    epyqlib.checkresultmodel.Result(
                        node=self,
                        severity=(epyqlib.checkresultmodel.ResultSeverity.information),
                        message=("Connected to old-style parameter"),
                    )
                )

                access_level = root.model.node_from_uuid(
                    parameter.access_level_uuid,
                )

                if access_level.value > 0:
                    result.append_child(
                        epyqlib.checkresultmodel.Result(
                            node=self,
                            severity=(epyqlib.checkresultmodel.ResultSeverity.warning),
                            message=("Access level will not be enforced"),
                        )
                    )

        return result

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move


@graham.schemify(tag="data_point_bitfield_member", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class DataPointBitfieldMember(epyqlib.treenode.TreeNode):
    parameter_uuid = create_parameter_uuid_attribute()

    bit_offset = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_int_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )

    bit_length = create_size_attribute(default=1)

    type_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Type",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root="sunspec types",
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, epyqlib.pm.parametermodel.Parameter)

    def child_from(self, node):
        self.parameter_uuid = node.uuid

        return None

    can_delete = epyqlib.attrsmodel.childless_can_delete
    all_addable_types = epyqlib.attrsmodel.empty_all_addable_types
    addable_types = epyqlib.attrsmodel.empty_addable_types
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="data_point_bitfield", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class DataPointBitfield(epyqlib.treenode.TreeNode):
    parameter_uuid = create_parameter_uuid_attribute()

    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(graham.schema(DataPointBitfieldMember)),
                )
            ),
        ),
    )

    # TODO: though this only really makes sense as one of the bitfield types
    type_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Type",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root="sunspec types",
    )

    block_offset = attr.ib(
        default=0,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )

    size = create_size_attribute()

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(
            node,
            (
                DataPointBitfieldMember,
                epyqlib.pm.parametermodel.Parameter,
            ),
        )

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    def child_from(self, node):
        if isinstance(node, epyqlib.pm.parametermodel.Parameter):
            self.parameter_uuid = node.uuid
            return None

        return node

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


def check_block_offsets_and_length(self):
    length = 0

    root = self.find_root()

    for point in self.children:
        try:
            type_ = root.model.node_from_uuid(point.type_uuid)
        except epyqlib.attrsmodel.NotFoundError:
            point_name = root.model.node_from_uuid(point.parameter_uuid).name
            raise TypeNotFoundError(
                f"Point {point_name!r}" f" has unknown type uuid {point.type_uuid!r}"
            )

        if type_.name != "string" and point.size != type_.value:
            point_name = root.model.node_from_uuid(point.parameter_uuid).name
            raise MismatchedSizeAndTypeError(
                f"Expected {type_.value} for {type_.name}"
                f", is {point.size} for {point_name}"
            )

        length += point.size

    return length


@graham.schemify(tag="sunspec_header_block", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class HeaderBlock(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="Header",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    offset = attr.ib(
        default=0,
        converter=int,
    )
    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(marshmallow.fields.Nested(graham.schema(DataPoint)),)
            ),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        return False

    check_offsets_and_length = check_block_offsets_and_length

    def add_data_points(self, uint16_uuid, model_id):
        parameters = [
            epyqlib.pm.parametermodel.Parameter(
                name=model_id,
                abbreviation="ID",
                read_only=True,
            ),
            epyqlib.pm.parametermodel.Parameter(
                name="",
                abbreviation="L",
                comment="Model Length",
                read_only=True,
            ),
        ]
        points = [
            DataPoint(
                block_offset=0,
                size=1,
                type_uuid=uint16_uuid,
                parameter_uuid=parameters[0].uuid,
                mandatory=True,
            ),
            DataPoint(
                block_offset=1,
                size=1,
                type_uuid=uint16_uuid,
                parameter_uuid=parameters[1].uuid,
                mandatory=True,
            ),
        ]

        for point in points:
            self.append_child(point)

        return parameters

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="sunspec_fixed_block", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class FixedBlock(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="Fixed Block",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    offset = attr.ib(
        default=2,
        converter=int,
    )
    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(graham.schema(DataPoint)),
                    marshmallow.fields.Nested(graham.schema(DataPointBitfield)),
                )
            ),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()
        self.pyqt_signals.child_added_complete.connect(self.update)
        self.pyqt_signals.child_removed_complete.connect(self.update)

    def can_drop_on(self, node):
        return isinstance(
            node,
            (
                epyqlib.pm.parametermodel.Parameter,
                DataPoint,
                DataPointBitfield,
            ),
        )

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return isinstance(node, (DataPoint, DataPointBitfield))

    def child_from(self, node):
        if isinstance(node, epyqlib.pm.parametermodel.Parameter):
            return DataPoint(parameter_uuid=node.uuid)

        return node

    def update(self):
        block_offset = 0
        for pt in self.children:
            pt.block_offset = block_offset
            block_offset = block_offset + pt.size

    check_offsets_and_length = check_block_offsets_and_length
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(
    tag="sunspec_table_repeating_block_reference_data_point_reference",
    register=True,
)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("parameter_uuid",),
)
@attr.s(hash=False)
class TableRepeatingBlockReferenceDataPointReference(epyqlib.treenode.TreeNode):
    parameter_uuid = create_parameter_uuid_attribute()

    factor_uuid = create_factor_uuid_attribute()
    original = epyqlib.attrsmodel.create_reference_attribute()

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    all_addable_types = epyqlib.attrsmodel.empty_all_addable_types
    addable_types = epyqlib.attrsmodel.empty_addable_types
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="sunspec_table_repeating_block", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("name",),
)
@attr.s(hash=False)
class TableRepeatingBlockReference(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="Table Repeating Block Reference",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    offset = attr.ib(
        default=2,
        converter=int,
    )
    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(
                        graham.schema(
                            TableRepeatingBlockReferenceDataPointReference,
                        )
                    ),
                )
            ),
        ),
    )

    original = epyqlib.attrsmodel.create_reference_attribute()

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return False

    def check_offsets_and_length(self):
        return self.original.check_block_offsets_and_length()

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="sunspec_table_repeating_block", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("name",),
)
@attr.s(hash=False)
class TableDataPointReference(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="Table Data Point Reference",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    offset = attr.ib(
        default=2,
        converter=int,
    )
    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=()),
        ),
    )

    original = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=epyqlib.attrsmodel.Reference(allow_none=True),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=original,
        no_column=True,
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        return False

    check_offsets_and_length = check_block_offsets_and_length
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from


@graham.schemify(tag="table_model_reference", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class TableRepeatingBlock(epyqlib.treenode.TreeNode):
    uuid = epyqlib.attrsmodel.attr_uuid()
    name = attr.ib(
        default="Table Reference",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    abbreviation = epyqlib.pm.parametermodel.create_abbreviation_attribute()

    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(marshmallow.fields.Nested(graham.schema(DataPoint)),)
            ),
        ),
    )

    offset = attr.ib(
        default=2,
        converter=int,
    )

    repeats = attr.ib(
        default=0,
        converter=int,
    )

    path = attr.ib(
        factory=tuple,
    )
    epyqlib.attrsmodel.attrib(
        attribute=path,
        no_column=True,
    )
    graham.attrib(
        attribute=path,
        field=graham.fields.Tuple(marshmallow.fields.UUID()),
    )

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return False

    def check_block_offsets_and_length(self):
        return self.repeats * check_block_offsets_and_length(self)

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="table", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Table(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="New Table",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    parameter_table_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=parameter_table_uuid,
        human_name="Table UUID",
    )

    children = attr.ib(
        default=attr.Factory(list),
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(graham.schema(TableRepeatingBlock)),
                    marshmallow.fields.Nested(graham.schema(DataPoint)),
                )
            ),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    @staticmethod
    def can_drop_on(node):
        return isinstance(node, epyqlib.pm.parametermodel.Table)

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    def child_from(self, node):
        self.parameter_table_uuid = node.uuid
        return None

    def update(self, table=None):
        old_nodes = self.recursively_remove_children()
        old_nodes_by_path = {
            getattr(node, "path", getattr(node, "parameter_uuid", node.uuid)): node
            for node in old_nodes
        }

        if self.parameter_table_uuid is None:
            return

        root = self.find_root()
        model = root.model

        if table is None:
            table = model.node_from_uuid(self.parameter_table_uuid)
        elif table.uuid != self.table_uuid:
            raise ConsistencyError()

        master_array_data_points_by_uuid = {}

        for section in table.arrays_and_groups:
            if isinstance(section, epyqlib.pm.parametermodel.Array):
                array_element = section.children[0]
                node = old_nodes_by_path.get(array_element.uuid)
                if node is None:
                    node = DataPoint(
                        parameter_uuid=array_element.uuid,
                    )
                self.append_child(node)
                master_array_data_points_by_uuid[array_element.uuid] = node
            elif isinstance(section, epyqlib.pm.parametermodel.Group):
                for element in section.children:
                    node = old_nodes_by_path.get(element.uuid)
                    if node is None:
                        node = DataPoint(
                            parameter_uuid=element.uuid,
                        )
                    self.append_child(node)
                    master_array_data_points_by_uuid[element.uuid] = node

        for combination in table.combinations:
            not_first_curve = any(
                (
                    layer.tree_parent.name == "Curves"
                    and layer.name != layer.tree_parent.children[0].name
                )
                for layer in combination
            )
            if not_first_curve:
                continue

            curve_enumeration_search = [
                layer for layer in combination if layer.tree_parent.name == "Curves"
            ]
            if len(curve_enumeration_search) == 0:
                curve_count = 0
            elif len(curve_enumeration_search) == 1:
                curve_enumeration = curve_enumeration_search[0].tree_parent
                curve_count = len(curve_enumeration.children)
            else:
                raise blue

            base_path = tuple(node.uuid for node in combination)

            block_node = old_nodes_by_path.get(base_path)

            if block_node is None:
                block_node = TableRepeatingBlock(
                    name=" - ".join(
                        item.name
                        for item in combination
                        if item.tree_parent.name != "Curves"
                    ),
                    path=base_path,
                )

            block_node.repeats = curve_count

            self.append_child(block_node)

            (in_tree,) = table.group.nodes_by_attribute(
                attribute_value=tuple(node.uuid for node in combination),
                attribute_name="path",
            )
            # continue

            block_offset = 0

            group_elements = [[], []]
            group_of_groups = group_elements[0]

            for child in in_tree.children:
                if isinstance(child.original, epyqlib.pm.parametermodel.Array):
                    group_of_groups = group_elements[1]
                    continue

                group_of_groups.append(child.children)

            for group in group_elements:
                group[:] = itertools.chain.from_iterable(group)

            # TODO: CAMPid 143707880547014313476753071297360068134
            for element in group_elements[0]:
                point_node = old_nodes_by_path.get(element.uuid)
                reference_data_point = master_array_data_points_by_uuid[
                    element.original.uuid
                ]
                if point_node is None:
                    point_node = DataPoint(
                        parameter_uuid=element.uuid,
                    )
                point_node.mandatory = reference_data_point.mandatory
                point_node.units = reference_data_point.units
                point_node.type_uuid = reference_data_point.type_uuid
                point_node.size = reference_data_point.size
                point_node.enumeration_uuid = reference_data_point.enumeration_uuid
                point_node.block_offset = block_offset
                block_node.append_child(point_node)
                block_offset += point_node.size

            array_elements = itertools.chain.from_iterable(
                zip(
                    *(
                        array.children
                        for array in in_tree.children
                        if isinstance(array.original, epyqlib.pm.parametermodel.Array)
                    )
                ),
            )
            for element in array_elements:
                point_node = old_nodes_by_path.get(element.uuid)
                reference_data_point = master_array_data_points_by_uuid[
                    element.tree_parent.children[0].original.uuid
                ]
                if point_node is None:
                    point_node = DataPoint(
                        parameter_uuid=element.uuid,
                    )
                point_node.mandatory = reference_data_point.mandatory
                point_node.units = reference_data_point.units
                point_node.type_uuid = reference_data_point.type_uuid
                point_node.size = reference_data_point.size
                point_node.enumeration_uuid = reference_data_point.enumeration_uuid
                point_node.block_offset = block_offset
                block_node.append_child(point_node)
                block_offset += point_node.size

            # TODO: CAMPid 143707880547014313476753071297360068134
            for element in group_elements[1]:
                point_node = old_nodes_by_path.get(element.uuid)
                reference_data_point = master_array_data_points_by_uuid[
                    element.original.uuid
                ]
                if point_node is None:
                    point_node = DataPoint(
                        parameter_uuid=element.uuid,
                    )
                point_node.mandatory = reference_data_point.mandatory
                point_node.units = reference_data_point.units
                point_node.type_uuid = reference_data_point.type_uuid
                point_node.size = reference_data_point.size
                point_node.enumeration_uuid = reference_data_point.enumeration_uuid
                point_node.block_offset = block_offset
                block_node.append_child(point_node)
                block_offset += point_node.size

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="sunspec_model", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Model(epyqlib.treenode.TreeNode):
    id = attr.ib(
        default=0,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )  # 103
    length = attr.ib(
        default=0,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )  # 50, in the spirit of over constraining like how
    # data points have their offset

    # ?
    # self.model_id = '{}'.format(model_id)
    # self.namespace = namespace
    # self.index = index

    # special children
    # header: id and length

    # self.point_data = []
    children = attr.ib(
        factory=lambda: [HeaderBlock(), FixedBlock()],
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(graham.schema(HeaderBlock)),
                    marshmallow.fields.Nested(graham.schema(FixedBlock)),
                    marshmallow.fields.Nested(
                        graham.schema(TableRepeatingBlockReference)
                    ),
                )
            ),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def remove_old_on_drop(self, node):
        return False

    def child_from(self, node):
        reference = TableRepeatingBlockReference(original=node)
        for child in node.tree_parent.children:
            if not isinstance(child, DataPoint):
                continue

            reference.append_child(
                TableRepeatingBlockReferenceDataPointReference(
                    original=child,
                ),
            )

        return reference

    def can_drop_on(self, node):
        return isinstance(node, TableRepeatingBlock) and not any(
            isinstance(child, TableRepeatingBlockReference) for child in self.children
        )

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return not isinstance(node, (HeaderBlock, FixedBlock))

    def check_offsets_and_length(self):
        length = 0

        for block in self.children:
            length += block.check_offsets_and_length()

        return length

    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


# class Repeating...?


Root = epyqlib.attrsmodel.Root(
    default_name="SunSpec",
    valid_types=(Model, Table),
)


types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        Model,
        DataPoint,
        DataPointBitfield,
        DataPointBitfieldMember,
        HeaderBlock,
        FixedBlock,
        Table,
        TableRepeatingBlock,
        TableRepeatingBlockReference,
        TableRepeatingBlockReferenceDataPointReference,
    ),
)


# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    (
        merge(
            "name",
            HeaderBlock,
            FixedBlock,
            Table,
            TableRepeatingBlock,
            TableRepeatingBlockReference,
        )
        + merge("id", Model)
        + merge(
            "parameter_uuid",
            DataPoint,
            TableRepeatingBlockReferenceDataPointReference,
            DataPointBitfield,
            DataPointBitfieldMember,
        )
    ),
    merge("abbreviation", TableRepeatingBlock),
    merge("not_implemented", DataPoint),
    merge("length", Model) + merge("size", DataPoint, DataPointBitfield),
    merge("repeats", TableRepeatingBlock),
    merge("hand_coded_getter", DataPoint),
    merge("hand_coded_setter", DataPoint),
    merge(
        "factor_uuid",
        DataPoint,
        TableRepeatingBlockReferenceDataPointReference,
    ),
    merge("units", DataPoint),
    merge("enumeration_uuid", DataPoint),
    merge("type_uuid", DataPoint, DataPointBitfield, DataPointBitfieldMember),
    merge("bit_length", DataPointBitfieldMember),
    merge("bit_offset", DataPointBitfieldMember),
    merge("parameter_table_uuid", Table),
    merge("mandatory", DataPoint),
    merge(
        "offset",
        DataPoint,
        HeaderBlock,
        FixedBlock,
        TableRepeatingBlockReference,
        TableRepeatingBlock,
    ),
    merge("block_offset", DataPoint, DataPointBitfield),
    merge("uuid", *types.types.values()),
)


# TODO: CAMPid 075454679961754906124539691347967
@attr.s
class ReferencedUuidNotifier:
    changed = epyqlib.utils.qt.Signal("PyQt_PyObject")

    view = attr.ib(default=None)
    selection_model = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.view is not None:
            self.set_view(self.view)

    def set_view(self, view):
        self.disconnect_view()

        self.view = view
        self.selection_model = self.view.selectionModel()
        self.selection_model.currentChanged.connect(
            self.current_changed,
        )

    def disconnect_view(self):
        if self.selection_model is not None:
            self.selection_model.currentChanged.disconnect(
                self.current_changed,
            )
        self.view = None
        self.selection_model = None

    def current_changed(self, current, previous):
        if not current.isValid():
            return

        index = epyqlib.utils.qt.resolve_index_to_model(
            index=current,
        )
        model = index.data(epyqlib.utils.qt.UserRoles.attrs_model)
        node = model.node_from_index(index)

        parameter_uuid = getattr(node, "parameter_uuid", None)

        if parameter_uuid is not None:
            self.changed.emit(parameter_uuid)



import click

import epyqlib.pm.parametermodel
import epcpm.parameterstodocx
import epcpm.project


@click.command()
@click.option("--project", "project_file", type=click.File(), required=True)
@click.option("--docx", "docx_file", type=click.File("wb"), required=True)
@click.option("--template", type=click.File("rb"))
@click.option("--access-level", default="user")
def cli(project_file, docx_file, template, access_level):
    project = epcpm.project.load(project_file)

    (access_levels,) = project.models.parameters.root.nodes_by_filter(
        filter=(lambda node: isinstance(node, epyqlib.pm.parametermodel.AccessLevels)),
    )

    access_level = access_levels.by_name(access_level)

    docx_builder = epcpm.parameterstodocx.builders.wrap(
        wrapped=project.models.parameters.root,
        can_root=project.models.can.root,
        template=template,
        access_level=access_level,
    )

    doc = docx_builder.gen()

    doc.save(docx_file)
